import sys
import time

from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from timeline import Timeline
from oops_timeline.hooks import flatten_timeline
from raven.contrib.django.raven_compat.models import client

from canonical_raven.middleware import RavenContextMiddleware, Sentry


class TestClient(object):

    def __init__(self, *args, **kwargs):
        self.reset()

    def http_context(self, data, **kwargs):
        pass

    def extra_context(self, data, **kwargs):
        self.context['extra'] = data

    def captureException(self, exc_info=None, **kwargs):
        self.exception_captured = True

    def reset(self):
        self.context = {}
        self.exception_captured = False


@override_settings(
    SENTRY_CLIENT='canonical_raven.tests.test_middleware.TestClient')
class TestClientTestCase(TestCase):

    def assert_extra_context(self, context, expected):
        self.assertIn('extra', context)
        extra = context['extra']
        for key, value in expected.items():
            self.assertEqual(extra[key], value)

    def assert_no_extra_context(self, context):
        self.assertNotIn('extra', context)


class RavenContextMiddlewareTestCase(TestClientTestCase):

    def setUp(self):
        super(RavenContextMiddlewareTestCase, self).setUp()
        self.middleware = RavenContextMiddleware()
        self.addCleanup(client.reset)

    def test_report_error_raven_includes_timeline_in_extra_context(self):
        timeline = Timeline()
        request = RequestFactory().get('/')
        request.environ['timeline.timeline'] = timeline
        expected = {}
        flatten_timeline(expected, {'timeline': timeline})

        self.middleware.process_exception(request, None)
        self.assert_extra_context(client.context, expected)

    def test_report_error_raven_includes_oops_id_in_extra_context(self):
        request = RequestFactory().get('/')
        request.environ['oops.report'] = {'id': 'oops-id-1234'}
        expected = {'X-Oops-Id': 'oops-id-1234'}

        self.middleware.process_exception(request, None)
        self.assert_extra_context(client.context, expected)


class SentryMiddlewareTestCase(TestClientTestCase):

    def setUp(self):
        super(SentryMiddlewareTestCase, self).setUp()
        self.soft_start_timeout = 1
        self.addCleanup(client.reset)

    def make_environ(self):
        return {
            'wsgi.url_scheme': 'http',
            'SERVER_NAME': 'localhost',
            'SERVER_PORT': '8000',
        }

    def make_middleware(self, timeout=None):
        def application(environ, start_response):
            time.sleep(0.001)
            start_response(200, {})
            return 'OK'
        return Sentry(application, soft_start_timeout=timeout)

    def dummy_start_response(self, status, headers, exc_info=None):
        self.response_status = status
        self.response_headers = headers
        self.response_exc_info = exc_info

    def test_handle_exception_includes_timeline_in_extra_context(self):
        timeline = Timeline()
        request = RequestFactory().get('/')
        request.environ['timeline.timeline'] = timeline
        expected = {}
        flatten_timeline(expected, {'timeline': timeline})

        middleware = self.make_middleware()
        middleware.handle_exception(request.environ)
        self.assert_extra_context(middleware.client.context, expected)

    def test_trigger_exception_on_soft_timeout(self):
        middleware = self.make_middleware(timeout=1)
        environ = self.make_environ()

        response = middleware(environ, self.dummy_start_response)
        # consume response
        response.next()

        # extra context was defined, but is empty due to raven's
        # default scrubbing policy
        self.assert_extra_context(middleware.client.context, {})
        self.assertTrue(middleware.client.exception_captured)

    def test_does_not_trigger_exception_when_timeout_not_exceeded(self):
        middleware = self.make_middleware(timeout=10)
        environ = self.make_environ()

        response = middleware(environ, self.dummy_start_response)
        # consume response
        response.next()

        self.assert_no_extra_context(middleware.client.context)
        self.assertFalse(middleware.client.exception_captured)

    def test_does_not_trigger_exception_when_no_timeout(self):
        middleware = self.make_middleware()
        environ = self.make_environ()

        response = middleware(environ, self.dummy_start_response)
        # consume response
        response.next()

        self.assert_no_extra_context(middleware.client.context)
        self.assertFalse(middleware.client.exception_captured)

    def test_handle_start_response_with_exc_info(self):
        try:
            raise ValueError('foo')
        except Exception:
            exc_info = sys.exc_info()

        def application(environ, start_response):
            start_response(500, {}, exc_info)
            return 'ERROR'

        middleware = Sentry(application)
        environ = self.make_environ()

        response = middleware(environ, self.dummy_start_response)
        # consume response
        response.next()

        self.assertEqual(self.response_status, 500)
        self.assertEqual(self.response_headers, {})
        self.assertEqual(self.response_exc_info, exc_info)
